@@ -17,6 +17,8 @@ pub use snow::Keypair;
 
 use crate::{Error, MAX_MESSAGE_LEN, TAG_LEN};
 
+const LENGTH_FIELD_LEN: usize = std::mem::size_of::<u16>();
+
 #[derive(Debug)]
 enum ReadState {
     ShuttingDown,
@@ -60,23 +62,38 @@ impl<T: Debug> Debug for NoiseStream<T> {
     }
 }
 
-impl<T> NoiseStream<T>
-where
-    T: AsyncRead + AsyncWrite + Unpin,
-{
+impl<T> NoiseStream<T> {
+    pub fn get_inner(&self) -> &T {
+        &self.inner
+    }
+
+    pub fn get_inner_mut(&mut self) -> &mut T {
+        &mut self.inner
+    }
+
     pub fn into_inner(self) -> T {
         self.inner
     }
 
-    pub fn state(&self) -> &TransportState {
+    pub fn get_state(&self) -> &TransportState {
         &self.transport
     }
 
-    pub fn state_mut(&mut self) -> &mut TransportState {
+    pub fn get_state_mut(&mut self) -> &mut TransportState {
         &mut self.transport
     }
+}
 
-    pub async fn handshake(mut inner: T, mut state: HandshakeState) -> Result<Self, Error> {
+impl<T> NoiseStream<T>
+where
+    T: AsyncRead + AsyncWrite + Unpin,
+{
+    pub async fn handshake_with_verifier<F: FnOnce(&[u8]) -> bool>(
+        mut inner: T,
+        mut state: HandshakeState,
+        verifier: F,
+    ) -> Result<Self, Error> {
+        let mut f = Some(verifier);
         loop {
             if state.is_handshake_finished() {
                 let transport = state.into_transport_mode()?;
@@ -86,9 +103,9 @@ where
                     read_state: ReadState::Idle,
                     write_state: WriteState::Idle,
                     write_clean_waker: None,
-                    read_message_buffer: vec![],
-                    read_payload_buffer: vec![],
-                    write_message_buffer: vec![],
+                    read_message_buffer: vec![0; MAX_MESSAGE_LEN],
+                    read_payload_buffer: vec![0; MAX_MESSAGE_LEN],
+                    write_message_buffer: vec![0; LENGTH_FIELD_LEN + MAX_MESSAGE_LEN],
                 });
             }
 
@@ -99,13 +116,26 @@ where
                 let len = state.write_message(&[], &mut message)?;
                 inner.write_u16_le(len as u16).await?;
                 inner.write_all(&message[..len]).await?;
+                inner.flush().await?;
             } else {
                 let len = inner.read_u16_le().await? as usize;
                 inner.read_exact(&mut message[..len]).await?;
                 state.read_message(&message[..len], &mut payload)?;
+                if let Some(pubkey) = state.get_remote_static() {
+                    if let Some(verifier) = f.take() {
+                        if !verifier(pubkey) {
+                            return Err(Error::HandshakeError("invalid public key".into()));
+                        }
+                    }
+                }
             }
         }
     }
+
+    #[inline]
+    pub async fn handshake(inner: T, state: HandshakeState) -> Result<Self, Error> {
+        Self::handshake_with_verifier(inner, state, |_| true).await
+    }
 }
 
 impl<T> AsyncWrite for NoiseStream<T>
@@ -131,12 +161,18 @@ where
                 WriteState::Idle => {
                     let payload_len = buf.len().min(MAX_MESSAGE_LEN - TAG_LEN);
                     let buf = &buf[..payload_len];
-                    write_message_buffer.resize(2 + MAX_MESSAGE_LEN, 0);
+
+                    // Safety: This is safe because this buffer is initialized with length LENGTH_FIELD_LEN + MAX_MESSAGE_LEN
+                    unsafe {
+                        write_message_buffer.set_len(LENGTH_FIELD_LEN + MAX_MESSAGE_LEN);
+                    }
+
                     let message_len = transport
-                        .write_message(buf, &mut write_message_buffer[2..])
+                        .write_message(buf, &mut write_message_buffer[LENGTH_FIELD_LEN..])
                         .map_err(|e| std::io::Error::new(ErrorKind::InvalidData, e))?;
-                    write_message_buffer[..2].copy_from_slice(&(message_len as u16).to_le_bytes());
-                    write_message_buffer.truncate(2 + message_len);
+                    write_message_buffer[..LENGTH_FIELD_LEN]
+                        .copy_from_slice(&(message_len as u16).to_le_bytes());
+                    write_message_buffer.truncate(LENGTH_FIELD_LEN + message_len);
                     *state = WriteState::WritingMessage(0, payload_len);
                 }
                 WriteState::WritingMessage(start, payload_len) => {
@@ -208,11 +244,15 @@ where
                 ReadState::ShuttingDown => {
                     return Poll::Ready(Ok(()));
                 }
-                ReadState::Idle => *state = ReadState::ReadingLen(0, [0; 2]),
+                ReadState::Idle => *state = ReadState::ReadingLen(0, [0; LENGTH_FIELD_LEN]),
                 ReadState::ReadingLen(read_len, mut buf) => {
-                    if *read_len == 2 {
-                        let message_len = u16::from_le_bytes(buf) as usize;
-                        read_message_buffer.resize(message_len, 0);
+                    if *read_len == LENGTH_FIELD_LEN {
+                        let message_len = u16::from_le_bytes(buf);
+
+                        // Safety: This is safe because message_len <= MAX_MESSAGE_LEN
+                        unsafe {
+                            read_message_buffer.set_len(message_len as usize);
+                        }
                         *state = ReadState::ReadingMessage(0);
                     } else {
                         let mut read_buf = ReadBuf::new(&mut buf);
@@ -230,7 +270,11 @@ where
                 }
                 ReadState::ReadingMessage(start) => {
                     if *start == read_message_buffer.len() {
-                        read_payload_buffer.resize(read_message_buffer.len(), 0);
+                        // Safety: This is safe because this buffer is initialized with MAX_MESSAGE_LEN
+                        unsafe {
+                            read_payload_buffer.set_len(MAX_MESSAGE_LEN);
+                        }
+
                         let n = transport
                             .read_message(read_message_buffer, read_payload_buffer)
                             .map_err(|e| std::io::Error::new(ErrorKind::InvalidData, e))?;
